@@ -7,25 +7,29 @@ from importlib import import_module
 
 import stdnum
 import stdnum.exceptions
-from sql import Null, Column, Literal
+from sql import Literal
+from sql.functions import CharLength
 from sql.operators import Equal
 
 from trytond.i18n import gettext
 from trytond.model import (
-    ModelView, ModelSQL, Model, UnionMixin, DeactivableMixin, sequence_ordered,
-    Exclude, fields)
-from trytond.pyson import Eval, If
-from trytond.transaction import Transaction
-from trytond.pool import Pool
-from trytond import backend
-from trytond.tools import lstrip_wildcard
-from trytond.tools.multivalue import migrate_property
+    DeactivableMixin, Exclude, Index, Model, ModelSQL, ModelView, UnionMixin,
+    fields, sequence_ordered)
 from trytond.modules.company.model import (
     CompanyMultiValueMixin, CompanyValueMixin)
+from trytond.pool import Pool
+from trytond.pyson import Eval, Get, If
+from trytond.tools import is_full_text, lstrip_wildcard
+from trytond.transaction import Transaction
+
+try:
+    from trytond.tools import barcode
+except ImportError:
+    barcode = None
+
 from .exceptions import InvalidIdentifierCode
 from .ir import price_decimal
 
-
 __all__ = ['price_digits', 'round_price', 'TemplateFunction']
 logger = logging.getLogger(__name__)
 
@@ -56,25 +60,26 @@ class Template(
     __name__ = "product.template"
     _order_name = 'rec_name'
     name = fields.Char(
-        "Name", size=None, required=True, translate=True, select=True)
+        "Name", size=None, required=True, translate=True)
     code_readonly = fields.Function(
         fields.Boolean("Code Readonly"), 'get_code_readonly')
     code = fields.Char(
-        "Code", select=True,
+        "Code", strip='leading',
         states={
             'readonly': Eval('code_readonly', False),
-            },
-        depends=['code_readonly'])
+            })
     type = fields.Selection(TYPES, "Type", required=True)
     consumable = fields.Boolean('Consumable',
         states={
             'invisible': Eval('type', 'goods') != 'goods',
             },
-        depends=['type'],
         help="Check to allow stock moves to be assigned "
         "regardless of stock level.")
     list_price = fields.MultiValue(fields.Numeric(
-            "List Price", required=True, digits=price_digits,
+            "List Price", digits=price_digits,
+            states={
+                'readonly': ~Eval('context', {}).get('company'),
+                },
             help="The standard price the product is sold at."))
     list_prices = fields.One2Many(
         'product.list_price', 'template', "List Prices")
@@ -88,16 +93,17 @@ class Template(
             help="The method used to calculate the cost price."))
     cost_price_methods = fields.One2Many(
         'product.cost_price_method', 'template', "Cost Price Methods")
-    default_uom = fields.Many2One('product.uom', "Default UOM", required=True,
-        help="The standard unit of measure for the product.\n"
+    default_uom = fields.Many2One(
+        'product.uom', "Default UoM", required=True,
+        help="The standard Unit of Measure for the product.\n"
         "Used internally when calculating the stock levels of goods "
         "and assets.")
     default_uom_category = fields.Function(
-        fields.Many2One('product.uom.category', 'Default UOM Category'),
+        fields.Many2One(
+            'product.uom.category', "Default UoM Category",
+            help="The category of the default Unit of Measure."),
         'on_change_with_default_uom_category',
         searcher='search_default_uom_category')
-    default_uom_digits = fields.Function(fields.Integer("Default Unit Digits"),
-        'on_change_with_default_uom_digits')
     categories = fields.Many2Many(
         'product.template-product.category', 'template', 'category',
         "Categories",
@@ -111,25 +117,26 @@ class Template(
         domain=[
             If(~Eval('active'), ('active', '=', False), ()),
             ],
-        depends=['active'],
         help="The different variants the product comes in.")
 
     @classmethod
-    def __register__(cls, module_name):
-        super(Template, cls).__register__(module_name)
+    def __setup__(cls):
+        cls.code.search_unaccented = False
+        super().__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.update({
+                Index(t, (t.code, Index.Similarity())),
+                })
+        cls._order.insert(0, ('rec_name', 'ASC'))
 
-        table = cls.__table_handler__(module_name)
-
-        # Migration from 3.8: rename category into categories
-        if table.column_exist('category'):
-            logger.warning(
-                'The column "category" on table "%s" must be dropped manually',
-                cls._table)
+        types_cost_method = cls._cost_price_method_domain_per_type()
+        cls.cost_price_method.domain = [
+            Get(types_cost_method, Eval('type'), []),
+            ]
 
     @classmethod
-    def __setup__(cls):
-        super().__setup__()
-        cls._order.insert(0, ('rec_name', 'ASC'))
+    def _cost_price_method_domain_per_type(cls):
+        return {'service': [('cost_price_method', '=', 'fixed')]}
 
     @classmethod
     def multivalue_model(cls, field):
@@ -140,10 +147,18 @@ class Template(
             return pool.get('product.cost_price_method')
         return super(Template, cls).multivalue_model(field)
 
+    @classmethod
+    def order_code(cls, tables):
+        table, _ = tables[None]
+        if cls.default_code_readonly():
+            return [CharLength(table.code), table.code]
+        else:
+            return [table.code]
+
     @classmethod
     def order_rec_name(cls, tables):
         table, _ = tables[None]
-        return [table.code, table.name]
+        return cls.order_code(tables) + [table.name]
 
     def get_rec_name(self, name):
         if self.code:
@@ -153,19 +168,19 @@ class Template(
 
     @classmethod
     def search_rec_name(cls, name, clause):
-        if clause[1].startswith('!') or clause[1].startswith('not '):
+        _, operator, operand, *extra = clause
+        if operator.startswith('!') or operator.startswith('not '):
             bool_op = 'AND'
         else:
             bool_op = 'OR'
-        code_value = clause[2]
-        if clause[1].endswith('like'):
-            code_value = lstrip_wildcard(clause[2])
+        code_value = operand
+        if operator.endswith('like') and is_full_text(operand):
+            code_value = lstrip_wildcard(operand)
         return [bool_op,
-            ('name',) + tuple(clause[1:]),
-            ('code', clause[1], code_value) + tuple(clause[3:]),
-            ('products.code', clause[1], code_value) + tuple(clause[3:]),
-            ('products.identifiers.code', clause[1], code_value)
-            + tuple(clause[3:]),
+            ('name', operator, operand, *extra),
+            ('code', operator, code_value, *extra),
+            ('products.code', operator, code_value, *extra),
+            ('products.identifiers.code', operator, code_value, *extra),
             ]
 
     @staticmethod
@@ -213,18 +228,11 @@ class Template(
 
     @fields.depends('default_uom')
     def on_change_with_default_uom_category(self, name=None):
-        if self.default_uom:
-            return self.default_uom.category.id
+        return self.default_uom.category if self.default_uom else None
 
     @classmethod
     def search_default_uom_category(cls, name, clause):
-        return [('default_uom.category' + clause[0].lstrip(name),)
-            + tuple(clause[1:])]
-
-    @fields.depends('default_uom')
-    def on_change_with_default_uom_digits(self, name=None):
-        if self.default_uom:
-            return self.default_uom.digits
+        return [('default_uom.category' + clause[0][len(name):], *clause[1:])]
 
     @classmethod
     def _new_code(cls):
@@ -321,12 +329,11 @@ class Product(
     _order_name = 'rec_name'
     template = fields.Many2One(
         'product.template', "Product Template",
-        required=True, ondelete='CASCADE', select=True,
+        required=True, ondelete='CASCADE',
         search_context={'default_products': False},
         domain=[
             If(Eval('active'), ('active', '=', True), ()),
             ],
-        depends=['active'],
         help="The product that defines the common properties "
         "inherited by the variant.")
     code_readonly = fields.Function(fields.Boolean('Code Readonly'),
@@ -338,19 +345,22 @@ class Product(
                 }),
         'on_change_with_prefix_code')
     suffix_code = fields.Char(
-        "Suffix Code",
+        "Suffix Code", strip='trailing',
         states={
             'readonly': Eval('code_readonly', False),
             },
-        depends=['code_readonly'],
         help="The unique identifier for the product (aka SKU).")
-    code = fields.Char("Code", readonly=True, select=True,
+    code = fields.Char(
+        "Code", readonly=True,
         help="A unique identifier for the variant.")
     identifiers = fields.One2Many(
         'product.identifier', 'product', "Identifiers",
         help="Other identifiers associated with the variant.")
     cost_price = fields.MultiValue(fields.Numeric(
-            "Cost Price", required=True, digits=price_digits,
+            "Cost Price", digits=price_digits,
+            states={
+                'readonly': ~Eval('context', {}).get('company'),
+                },
             help="The amount it costs to purchase or make the variant, "
             "or carry out the service."))
     cost_prices = fields.One2Many(
@@ -370,6 +380,9 @@ class Product(
             cls._no_template_field = set()
         cls._no_template_field.update(['products'])
 
+        cls.suffix_code.search_unaccented = False
+        cls.code.search_unaccented = False
+
         super(Product, cls).__setup__()
         cls.__access__.add('template')
         cls._order.insert(0, ('rec_name', 'ASC'))
@@ -381,6 +394,8 @@ class Product(
                     & (t.code != '')),
                 'product.msg_product_code_unique'),
             ]
+        cls._sql_indexes.add(
+            Index(t, (t.code, Index.Similarity())))
 
         for attr in dir(Template):
             tfield = getattr(Template, attr)
@@ -398,8 +413,6 @@ class Product(
                     tfield.states['invisible'] |= invisible_state
                 else:
                     tfield.states['invisible'] = invisible_state
-                if 'template' not in tfield.depends:
-                    tfield.depends.append('template')
                 setattr(cls, attr, TemplateFunction(tfield))
                 order_method = getattr(cls, 'order_%s' % attr, None)
                 if (not order_method
@@ -433,6 +446,24 @@ class Product(
         # Prevent NotImplementedError for One2Many
         pass
 
+    @classmethod
+    def order_suffix_code(cls, tables):
+        table, _ = tables[None]
+        if cls.default_code_readonly():
+            return [CharLength(table.suffix_code), table.suffix_code]
+        else:
+            return [table.suffix_code]
+
+    @classmethod
+    def order_code(cls, tables):
+        pool = Pool()
+        Template = pool.get('product.template')
+        table, _ = tables[None]
+        if cls.default_code_readonly() or Template.default_code_readonly():
+            return [CharLength(table.code), table.code]
+        else:
+            return [table.code]
+
     @fields.depends('template', '_parent_template.id')
     def on_change_template(self):
         for name, field in self._fields.items():
@@ -468,9 +499,24 @@ class Product(
             return pool.get('product.cost_price')
         return super(Product, cls).multivalue_model(field)
 
+    def set_multivalue(self, name, value, save=True, **pattern):
+        context = Transaction().context
+        if name in {'cost_price', 'list_price'} and not value:
+            if not pattern.get('company', context.get('company')):
+                return []
+        return super().set_multivalue(name, value, save=save, **pattern)
+
+    def get_multivalue(self, name, **pattern):
+        if isinstance(self._fields[name], TemplateFunction):
+            return self.template.get_multivalue(name, **pattern)
+        else:
+            return super().get_multivalue(name, **pattern)
+
     @classmethod
     def default_cost_price(cls, **pattern):
-        return Decimal(0)
+        context = Transaction().context
+        if pattern.get('company', context.get('company')):
+            return Decimal(0)
 
     @classmethod
     def search_template(cls, name, clause):
@@ -488,7 +534,7 @@ class Product(
                 }
         else:
             template = tables['template']
-        return [product.code] + Template.name.convert_order('name',
+        return cls.order_code(tables) + Template.name.convert_order('name',
             tables['template'], Template)
 
     def get_rec_name(self, name):
@@ -499,18 +545,19 @@ class Product(
 
     @classmethod
     def search_rec_name(cls, name, clause):
-        if clause[1].startswith('!') or clause[1].startswith('not '):
+        _, operator, operand, *extra = clause
+        if operator.startswith('!') or operator.startswith('not '):
             bool_op = 'AND'
         else:
             bool_op = 'OR'
-        code_value = clause[2]
-        if clause[1].endswith('like'):
-            code_value = lstrip_wildcard(clause[2])
+        code_value = operand
+        if operator.endswith('like') and is_full_text(operand):
+            code_value = lstrip_wildcard(operand)
         return [bool_op,
-            ('code', clause[1], code_value) + tuple(clause[3:]),
-            ('identifiers.code', clause[1], code_value) + tuple(clause[3:]),
-            ('template.name',) + tuple(clause[1:]),
-            ('template.code', clause[1], code_value) + tuple(clause[3:]),
+            ('code', operator, code_value, *extra),
+            ('identifiers.code', operator, code_value, *extra),
+            ('template.name', operator, operand, *extra),
+            ('template.code', operator, code_value, *extra),
             ]
 
     @staticmethod
@@ -547,6 +594,14 @@ class Product(
     def get_code_readonly(self, name):
         return self.default_code_readonly()
 
+    def identifier_get(self, types=None):
+        "Return the first identifier for the given types"
+        if isinstance(types, str) or types is None:
+            types = {types}
+        for identifier in self.identifiers:
+            if identifier.type in types:
+                return identifier
+
     @classmethod
     def _new_suffix_code(cls):
         pool = Pool()
@@ -594,6 +649,13 @@ class Product(
         for product in products:
             code = ''.join(filter(None, [
                         product.prefix_code, product.suffix_code]))
+            if cls.code.strip:
+                if cls.code.strip == 'leading':
+                    code = code.lstrip()
+                elif cls.code.strip == 'trailing':
+                    code = code.rstrip()
+                else:
+                    code = code.strip()
             if not code:
                 code = None
             if code != product.code:
@@ -605,82 +667,31 @@ class ProductListPrice(ModelSQL, CompanyValueMixin):
     "Product List Price"
     __name__ = 'product.list_price'
     template = fields.Many2One(
-        'product.template', "Template", ondelete='CASCADE', select=True,
+        'product.template', "Template", ondelete='CASCADE',
         context={
             'company': Eval('company', -1),
             },
-        depends=['company'])
+        depends={'company'})
     list_price = fields.Numeric("List Price", digits=price_digits)
 
     @classmethod
-    def __register__(cls, module_name):
-        exist = backend.TableHandler.table_exist(cls._table)
-
-        super(ProductListPrice, cls).__register__(module_name)
-
-        if not exist:
-            cls._migrate_property([], [], [])
-
-    @classmethod
-    def _migrate_property(cls, field_names, value_names, fields):
-        field_names.append('list_price')
-        value_names.append('list_price')
-        fields.append('company')
-        migrate_property(
-            'product.template', field_names, cls, value_names,
-            parent='template', fields=fields)
+    def __setup__(cls):
+        super().__setup__()
+        cls.company.required = True
 
 
 class ProductCostPriceMethod(ModelSQL, CompanyValueMixin):
     "Product Cost Price Method"
     __name__ = 'product.cost_price_method'
     template = fields.Many2One(
-        'product.template', "Template", ondelete='CASCADE', select=True,
+        'product.template', "Template", ondelete='CASCADE',
         context={
             'company': Eval('company', -1),
             },
-        depends=['company'])
+        depends={'company'})
     cost_price_method = fields.Selection(
         'get_cost_price_methods', "Cost Price Method")
 
-    @classmethod
-    def __register__(cls, module_name):
-        pool = Pool()
-        ProductCostPrice = pool.get('product.cost_price')
-        sql_table = cls.__table__()
-        cost_price = ProductCostPrice.__table__()
-        cursor = Transaction().connection.cursor()
-
-        exist = backend.TableHandler.table_exist(cls._table)
-        cost_price_exist = backend.TableHandler.table_exist(
-            ProductCostPrice._table)
-
-        super(ProductCostPriceMethod, cls).__register__(module_name)
-
-        # Migrate from 4.4: move cost_price_method from ProductCostPrice
-        if not exist and not cost_price_exist:
-            cls._migrate_property([], [], [])
-        elif not exist and cost_price_exist:
-            cost_price_table = backend.TableHandler(
-                ProductCostPrice, module_name)
-            if cost_price_table.column_exist('template'):
-                columns = ['create_uid', 'create_date',
-                    'write_uid', 'write_date',
-                    'template', 'cost_price_method']
-                cursor.execute(*sql_table.insert(
-                        columns=[Column(sql_table, c) for c in columns],
-                        values=cost_price.select(
-                            *[Column(cost_price, c) for c in columns])))
-
-    @classmethod
-    def _migrate_property(cls, field_names, value_names, fields):
-        field_names.append('cost_price_method')
-        value_names.append('cost_price_method')
-        fields.append('company')
-        migrate_property(
-            'product.template', field_names, cls, value_names,
-            parent='template', fields=fields)
-
     @classmethod
     def get_cost_price_methods(cls):
         pool = Pool()
@@ -695,68 +706,27 @@ class ProductCostPrice(ModelSQL, CompanyValueMixin):
     "Product Cost Price"
     __name__ = 'product.cost_price'
     product = fields.Many2One(
-        'product.product', "Product", ondelete='CASCADE', select=True,
+        'product.product', "Product", ondelete='CASCADE',
         context={
             'company': Eval('company', -1),
             },
-        depends=['company'])
+        depends={'company'})
     cost_price = fields.Numeric(
-        "Cost Price", digits=price_digits)
+        "Cost Price", required=True, digits=price_digits)
 
     @classmethod
-    def __register__(cls, module_name):
-        pool = Pool()
-        Product = pool.get('product.product')
-        sql_table = cls.__table__()
-        product = Product.__table__()
-        cursor = Transaction().connection.cursor()
-
-        exist = backend.TableHandler.table_exist(cls._table)
-
-        super(ProductCostPrice, cls).__register__(module_name)
-
-        table = cls.__table_handler__(module_name)
-        if not exist:
-            # Create template column for property migration
-            table.add_column('template', 'INTEGER')
-            cls._migrate_property([], [], [])
-
-        # Migration from 4.4: replace template by product
-        if table.column_exist('template'):
-            columns = ['create_uid', 'create_date',
-                'write_uid', 'write_date', 'cost_price']
-            cursor.execute(*sql_table.insert(
-                    columns=[Column(sql_table, c) for c in columns]
-                    + [sql_table.product],
-                    values=sql_table.join(product,
-                        condition=sql_table.template == product.template
-                        ).select(
-                        *[Column(sql_table, c) for c in columns]
-                        + [product.id],
-                        where=(sql_table.template != Null)
-                        & (sql_table.product == Null))))
-            cursor.execute(*sql_table.delete(
-                    where=(sql_table.template != Null)
-                    & (sql_table.product == Null)))
-            table.drop_column('template')
-
-    @classmethod
-    def _migrate_property(cls, field_names, value_names, fields):
-        field_names.append('cost_price')
-        value_names.append('cost_price')
-        fields.append('company')
-        migrate_property(
-            'product.template', field_names, cls, value_names,
-            parent='template', fields=fields)
+    def __setup__(cls):
+        super().__setup__()
+        cls.company.required = True
 
 
 class TemplateCategory(ModelSQL):
     'Template - Category'
     __name__ = 'product.template-product.category'
-    template = fields.Many2One('product.template', 'Template',
-        ondelete='CASCADE', required=True, select=True)
-    category = fields.Many2One('product.category', 'Category',
-        ondelete='CASCADE', required=True, select=True)
+    template = fields.Many2One(
+        'product.template', "Template", ondelete='CASCADE', required=True)
+    category = fields.Many2One(
+        'product.category', "Category", ondelete='CASCADE', required=True)
 
 
 class TemplateCategoryAll(UnionMixin, ModelSQL):
@@ -774,8 +744,8 @@ class ProductIdentifier(sequence_ordered(), ModelSQL, ModelView):
     "Product Identifier"
     __name__ = 'product.identifier'
     _rec_name = 'code'
-    product = fields.Many2One('product.product', "Product", ondelete='CASCADE',
-        required=True, select=True,
+    product = fields.Many2One(
+        'product.product', "Product", ondelete='CASCADE', required=True,
         help="The product identified by the code.")
     type = fields.Selection([
             (None, ''),
@@ -791,8 +761,16 @@ class ProductIdentifier(sequence_ordered(), ModelSQL, ModelView):
 
     @classmethod
     def __setup__(cls):
+        cls.code.search_unaccented = False
         super().__setup__()
         cls.__access__.add('product')
+        t = cls.__table__()
+        cls._sql_indexes.update({
+                Index(
+                    t,
+                    (t.product, Index.Equality()),
+                    (t.code, Index.Similarity())),
+                })
 
     @fields.depends('type', 'code')
     def on_change_with_code(self):
@@ -827,3 +805,8 @@ class ProductIdentifier(sequence_ordered(), ModelSQL, ModelView):
                         type=self.type_string,
                         code=self.code,
                         product=product))
+
+    def barcode(self, format='svg', **options):
+        if barcode and self.type in barcode.BARCODES:
+            generator = getattr(barcode, 'generate_%s' % format)
+            return generator(self.type, self.on_change_with_code(), **options)
